/*
 * Copyright (c) 2003, 2004, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.strobel.assembler.metadata.signatures;

import com.strobel.core.StringUtilities;

public final class Wildcard implements TypeArgument {
    private final FieldTypeSignature _superBound;
    private final FieldTypeSignature _extendsBound;

    private Wildcard(final FieldTypeSignature superBound, final FieldTypeSignature extendsBound) {
        _superBound = superBound;
        _extendsBound = extendsBound;
    }

    public static Wildcard make(final FieldTypeSignature superBound, final FieldTypeSignature extendsBound) {
        return new Wildcard(superBound, extendsBound);
    }

    public boolean isUnbounded() {
        return !hasSuperBound() && !hasExtendsBound();
    }

    public boolean hasSuperBound() {
        return _superBound != null &&
               _superBound != BottomSignature.make();
    }

    public boolean hasExtendsBound() {
        return _extendsBound != null &&
               !(_extendsBound instanceof SimpleClassTypeSignature &&
                 StringUtilities.equals("java.lang.Object", ((SimpleClassTypeSignature) _extendsBound).getName()));
    }

    public FieldTypeSignature getSuperBound() {
        return _superBound;
    }

    public FieldTypeSignature getExtendsBound() {
        return _extendsBound;
    }

    public void accept(final TypeTreeVisitor<?> v) {
        v.visitWildcard(this);
    }
}
